![ASIX](https://www.asix.com.tw/upload/2019_12_061/20191206160042ixfpyg5Wz1.jpg)

# [ASIX USB Ethernet ICs](https://www.asix.com.tw/en/product/USBEthernet)
* AX88179A/772D
* AX88179/178A

## Prerequisites

Prepare to build the driver, you need the Linux kernel sources installed on the
build machine, and make sure that the version of the running kernel must match
the installed kernel sources. If you don't have the kernel sources, you can get
it from [The Linux Kernel Archives](www.kernel.org) or contact to your Linux 
distributor. If you don't know how to do, please refer to KERNEL-HOWTO.

**Note**: Please make sure the kernel is built with one of the "Support for 
Host-side, XHCI, EHCI, OHCI, or UHCI" option support.

## Getting Start

1. Extract the compressed driver source file to your temporary directory by the
following command:
``` shell=
	$tar -xf DRIVER_SOURCE_PACKAGE.tar.bz2
```

2. Now, the driver source files should be extracted under the current directory.
Issue the following command to compile the driver:
```shell=
	$make
```

3. If the compilation is done, the **ax_usb_nic.ko** will be created under the
current directory.

## Usage

### Install driver to your system
* If you want to use modprobe command to mount the driver, issue the
following command to install the driver into your system:
```shell=
	$sudo make install
```
**Note**: This command will backup the built-in ax88179_178a driver if it exists.

* If you want to load the driver by modprobe command, issue the following 
commands:
```shell=
	$sudo modprobe ax_usb_nic 
```

* If you want to unload the driver by modprobe command, issue the following 
commands:
```shell=
	$sudo modprobe -r ax_usb_nic 
```

* If you want to check the information of driver, issue the following commands:
```shell=
	$modinfo ax_usb_nic 
```

### Install driver manually
* If you want to load the driver manually, go to the driver directory and
execute the following commands:
```shell=
	$sudo modprobe mii
	$sudo insmod ax_usb_nic.ko
```
* If you want to unload the driver, just executing the following command:
```shell=
	$sudo rmmod ax_usb_nic
```

* If you want to check the information of driver, issue the following commands:
```shell=
	$modinfo ax_usb_nic.ko
```

-------------------------------------------------------------------------------
# Programmer
ASIX USB Ethernet Linux Command Line Programming Tool

## Getting Start

1. Extract the compressed driver source file to your temporary directory by the
following command:
``` shell=
	$tar -xf DRIVER_SOURCE_PACKAGE.tar.bz2
```

2. Now, the driver source files should be extracted under the current directory.
Executing the following command to compile the driver:
```shell=
	$make
```
			
3. Load the driver manually, go to the driver directory and execute the 
following commands:
```shell=
	$sudo modprobe mii
	$sudo insmod ax_usb_nic.ko
```

4. If the compilation is well, the ioctl will be created under the current 
directory.

**Note**: The default way to find the interface is to scan the ASIX device using
 the ethx (x: 0~255).It is defined in the file, **command.h**.
```c=
	(As follows)
	...
	// DEFAULT_SCAN   : scan "eth0" - "eth255"
	// INTERFACE_SCAN : scan all available network interfaces
	#define NET_INTERFACE	DEFAULT_SCAN
	#define	DEFAULT_SCAN	0x00
	#define	INTERFACE_SCAN	0x01
	...
```
Adjust the contents of #define NET_INTERFACE to select the method you want.

## AX88179/178A EEPROM/eFuse Programmer

1. If you want to read out values of the EEPROM/EFUSE to a file, go to the 
driver directory and execute the following command:
```shell=
	$sudo ./ax88179_programmer reeprom 0 eeprom 512
	$sudo ./ax88179_programmer reeprom 1 efuse 64
```

2. If you want to write values of a file to the EEPROM/EFUSE,  go to the driver
directory and execute the following command:
```shell=
	$sudo ./ax88179_programmer weeprom 0 eeprom 512
	$sudo ./ax88179_programmer weeprom 1 efuse 64
```

3. If you want to change the MAC address of a dongle, go to the driver directory
 and execute the following command:
```shell=
	$sudo ./ax88179_programmer chgmac 0 mac_addr 512
	$sudo ./ax88179_programmer chgmac 1 mac_addr 64
```

4. If you need more information about the instructions, go to the driver 
directory and execute the following commands:
```shell=
	$sudo ./ax88179_programmer reeprom help
	AX88179/AX88178A Linux EEPROM/eFuse Programming Tool vX.X.X
	./ax88179_programmer reeprom [type] [file] [size]
	    -- AX88179_178A EEPROM/eFuse read tool
	        [type]    - 0: EEPROM,  1: eFuse
	        [file]    - Output file
	        [size]    - EEPROM/eFuse SIZE (bytes). EEPROM maximum 512 bytes,
			    eFuse maximum 64 bytes.

	$sudo ./ax88179_programmer weeprom help
	AX88179/AX88178A Linux EEPROM/eFuse Programming Tool vX.X.X
	./ax88179_programmer weeeprom [type] [file] [size]
	    -- AX88179_178A EEPROM/eFuse write tool
	        [type]    - 0: EEPROM,  1: eFuse
	        [file]    - Input file
	        [size]    - EEPROM/eFuse SIZE (bytes). EEPROM size 12-512 bytes,
			    eFuse maximum 64 bytes.

	$sudo ./ax88179_programmer chgmac help
	AX88179/AX88178A Linux EEPROM/eFuse Programming Tool vX.X.X
	./ax88179_programmer chgmac [type] [mac_addr] [size]
	    -- AX88179_178A EEPROM/eFuse write tool (specify MAC address)
	        [type]    - 0: EEPROM,  1: eFuse
	        [mac_addr]- MAC address (xx:xx:xx:xx:xx:xx)
	        [size]    - EEPROM/eFuse SIZE (bytes). EEPROM size 12-512 bytes,
			    eFuse maximum 64 bytes.
```

## AX88179A/772d Flash/eFuse Programmer
1. If you want to get help message for specific command, go to the driver 
directory and execute the following command:
```shell=
	$sudo ./ax88179a_programmer help [command]

#Example:
	$sudo ./ax88179a_programmer help refuse
	AX88179A/AX88772D Linux Flash/eFuse Programming Tool vX.X.X
	./ax88179a_programmer refuse -f [File]
	    -- AX88179A_772D Read eFuse
	        -f [File]   - eFuse file path
```

2. If you want to get the version of firmware, go to the driver directory 
and execute the following command:
```shell=
	$sudo ./ax88179a_programmer rversion

#Example:
	$sudo ./ax88179a_programmer rversion
	AX88179A/AX88772D Linux Flash/eFuse Programming Tool vX.X.X
	Firmware Version: vX.X.X
	SUCCESS
```

3. If you want to get the MAC address, go to the driver directory and execute 
the following command:
```shell=
	$sudo ./ax88179a_programmer rmacaddr

#Example:
	$sudo ./ax88179a_programmer rmacaddr
	AX88179A/AX88772D Linux Flash/eFuse Programming Tool vX.X.X
	MAC address: XX:XX:XX:XX:XX:XX
	SUCCESS
```

4. If you want to write values of a file to the flash, go to the driver 
directory and execute the following command:
```shell=
	$sudo ./ax88179a_programmer wflash [file]

#Example:
	$sudo ./ax88179a_programmer wflash flash.txt
	AX88179A/AX88772D Linux Flash/eFuse Programming Tool vX.X.X
	File FW Version: vX.X.X
	SUCCESS
```

5. If you want to write the eFuse, go to the driver directory and execute 
the following command:
```shell=
	$sudo ./ax88179a_programmer wefuse -m [MAC] -s [SN] -f [File]

#Example:
	#Change MAC address
	$sudo ./ax88179a_programmer wefuse -m XX:XX:XX:XX:XX:XX -f efuse.txt
	AX88179A/AX88772D Linux Flash/eFuse Programming Tool vX.X.X
	SUCCESS

	#Change Serial number
	$sudo ./ax88179a_programmer wefuse -s 1234568790 -f efuse.txt
	AX88179A/AX88772D Linux Flash/eFuse Programming Tool vX.X.X
	SUCCESS

	#Change MAC address & Serial number
	$sudo ./ax88179a_programmer wefuse -m XX:XX:XX:XX:XX:XX -s 12 -f efuse.txt
	AX88179A/AX88772D Linux Flash/eFuse Programming Tool vX.X.X
	SUCCESS
```

6. If you want to read the eFuse, go to the driver directory and execute 
the following command:
```shell=
	$sudo ./ax88179a_programmer refuse -f [File]

#Example:
	$sudo ./ax88179a_programmer refuse -f dump.txt
	AX88179A/AX88772D Linux Flash/eFuse Programming Tool vX.X.X
	SUCCESS
```

7. If you want to reload the flash or eFuse to check version or MAC address 
and so, go to the driver directory and execute the following command:
```shell=
	$sudo ./ax88179a_programmer reload

#Example:
	$sudo ./ax88179a_programmer reload
	AX88179A/AX88772D Linux Flash/eFuse Programming Tool vX.X.X
	SUCCESS
```
-------------------------------------------------------------------------------
[ASIX Electronics Corporation](https://www.asix.com.tw/)